#!/usr/local/bin/ruby
#
# Sazaby - a simple bbs for ruby beginers
#
# $Id$
#
# "This file is written in UTF-8N." '　'
#
$kcode = 'utf-8'
#require 'web'
require 'web/common'
require 'web/cookie'
require 'web/request'
require 'web/response'
#require 'web/persistent'
require 'web/interface'
require 'yaml/store'

class Sazaby
  
  RegexpUtf8Char = /(?:
		     [\x00-\x7f]                                     |
		     [\xc2-\xdf] [\x80-\xbf]                         |
		     \xe0        [\xa0-\xbf] [\x80-\xbf]             |
		     [\xe1-\xef] [\x80-\xbf] [\x80-\xbf]             |
		     \xf0        [\x90-\xbf] [\x80-\xbf] [\x80-\xbf] |
		     [\xf1-\xf3] [\x80-\xbf] [\x80-\xbf] [\x80-\xbf] |
		     \xf4        [\x80-\x8f] [\x80-\xbf] [\x80-\xbf]
		    )/nx
  
  # -------------------------
  # Constructor
  #
  def initialize(req, rsp)
    @req = req
    @rsp = rsp
    @log = nil
    @MaxPages = 20
    @MaxLength = 200
    @ThreadsPerPage = 5
    @Title = 'Sazaby'
    @PageTitle = 'a simple bbs for ruby beginers'
    @RDFFileName = 'index.rdf'
    @CurrentPage
    @Program = $0[/[^\/\\:]+$/o]
    server = ENV['HTTP_HOST'] || ENV['SERVER_NAME']
#    path = ENV['REQUEST_URI'][/((?:\/[^?\/]*)*)\//o]
#    path = ENV['SCRIPT_NAME'][0, ENV['SCRIPT_NAME'].rindex('/')] if path.nil?
#    path[/\/\/[^\/]+/o] = '' if path[0,2] == '//'
#    path.chop! if path[-1] == ?/
    @URI = 'http://%s%s' % [ server, @req.script_name ]
  end
  
  
  # -------------------------
  # View
  #
  def view
    self.readlog if @log.nil?
    @rsp.content_type = 'text/html'
    @rsp.write( <<"EOS" )
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
<html xmlns="http://www.w3.org/1999/xhtml" xml:lang="ja-JP" lang="ja-JP">
<head>
<meta http-equiv="Content-Type" content="text/html; charset=utf-8" />
<title>#@Title</title>
<link rel="stylesheet" type="text/css" href="sazaby.css" />
<link rel="alternate" type="application/rss+xml" title="RSS" href="index.rdf" />
</style>
</head>

<body>
<h1 class="title">#@Title</h1>
<h2 class="subtitle">the simple bbs for ruby</h2>

<form id="postThread" action="#@Program" method="post">
<div class="WritingForm">
  <h2>writing form</h2>
  <div class="kv">
    <div class="key"><label accesskey="t" for="inputSubject">subject:</label></div>
    <div class="value"><input id="inputSubject" name="inputSubject" type="text" value=""></div>
  </div>
  <div class="kv">
    <div class="key"><label accesskey="n" for="inputName">name:</label></div>
    <div class="value"><input id="inputName" name="inputName" type="text" value=""></div>
  </div>
  
  <div class="kv">
    <div class="key"><label accesskey="b" for="inputBody">body:</label></div>
    <div class="value"><textarea id="inputBody" name="inputBody"></textarea></div>
  </div>
  
  <p class="submit"><input id="submit" name="submit" type="submit" value="Submit">
    <input id="mode" name="mode" type="hidden" value="postThread"></p>
</div>
</form>
EOS
    
    #
    # Pager
    #
    page = pager( @req['page'] || 1 )
    @CurrentPage = page['current']
    later = nil
    older = nil
    if page['later'] then
      later = '<a href="%s?page=%d">&lt;&lt;</a>' % [ @Program, page['later'] ]
    else
      later = '<a>Latest</a>'
    end
    if page['older'] then
      older = '<a href="%s?page=%d">&gt;&gt;</a>' % [ @Program, page['older'] ]
    else
      older = '<a>Oldest</a>'
    end
    pagelist = page['pagelist'].collect do |x|
      if x == @CurrentPage then
	'<strong>%d</strong>' % [ x, x ]
      else
	'<a href="%s?page=%d">%d</a>' % [ @Program, x, x ]
      end
    end
    thisThreads = @log.reverse[((@CurrentPage-1)*@ThreadsPerPage), @ThreadsPerPage]
    
    
    @rsp.write( <<"EOS" )
<div class="Pager">
  <div class="later">#{later}</div>
  <div class="older">#{older}</div>
  <div class="pagelist">#{pagelist.join(' ')}</div>
</div>
EOS
    
    
    thisThreads.each do |thread|
      target = @log.index(thread) + 1
      @rsp.write( <<"EOS" )
<div class="Thread">
  <h2>#{target} #{thread['subject']}</h2>
EOS
      countJ = -1
      thread['children'].each do |article|
	countJ += 1
	@rsp.write( <<"EOS" )
  <div class="#{ countJ == 0 ? "Root" : "Child" }Article">
    <p class="body">#{article['body'].gsub("\n",'<br>')}</p>
    <p class="info">
      by <span class="name">#{article['name']}</span>
      at <span class="date">#{article['time'].strftime('%Y-%m-%d %H:%M')}</span>
    </p>
  </div>
EOS
      end
      @rsp.write( <<"EOS" )
  <form id="postReply" action="#@Program" method="post">
    <div class="ReplyForm">
      <h3>replying form</h3>
    <div class="kv">
      <div class="key"><label accesskey="t" for="inputSubject">subject:</label></div>
      <div class="value"><input id="inputSubject" name="inputSubject" type="text" value=""></div>
    </div>
    <div class="kv">
      <div class="key"><label accesskey="n" for="inputName">name:</label></div>
      <div class="value"><input id="inputName" name="inputName" type="text" value=""></div>
    </div>
    
    <div class="kv">
      <div class="key"><label accesskey="b" for="inputBody">body:</label></div>
      <div class="value"><textarea id="inputBody" name="inputBody"></textarea></div>
    </div>
    
    <p class="submit"><input id="submit" name="submit" type="submit" value="Submit">
      <input id="target" name="target" type="hidden" value="#{target}">
      <input id="mode" name="mode" type="hidden" value="postArticle"></p>
  </div>
  </form>
</div>
EOS
      target -= 1
    end
    
    @rsp.write( <<"EOS" )
<!-- Pager -->

<p class="SazabyLabel" title="#{Process.times.to_yaml}">- Airemix Sazaby 1.0 beta -</p>
</body>
</html>
EOS
  end
  
  
  # -------------------------
  # Post New Thread
  #
  def postThread
    thread = Hash.new
    thread['subject'] = @req['inputSubject']
    thread['LastModified'] = Time.now
    thread['children'] = Array.new
    article = Hash.new
    article['name'] = @req['inputName']
    article['body'] = @req['inputBody'].gsub("\x0D\x0A","\n").tr("\r","\n")
    article['time'] = Time.now
    thread['children'].push article
    db = YAML::Store.new('./yaml_db')
    db.transaction do
      @log = db['log']
      @log = Array.new if @log.nil?
      @log.push thread
    end
  end
  
  
  # -------------------------
  # Post New Article
  #
  def postArticle
    article = Hash.new
    article['name'] = @req['inputName']
    article['body'] = @req['inputBody']
    article['time'] = Time.now
    db = YAML::Store.new('./yaml_db')
    db.transaction do
      @log = db['log']
      @log = Array.new if @log.nil?
      target = @log
      @req['target'].split('/').each do |i|
	if target.nil? then break end
	if target.is_a?(Hash) then target = target['children'] end
	if target.is_a?(Array) then target = target[i.to_i - 1] end
	if target.nil? then break end
      end
      target['children'].push article
      target['LastModified'] = Time.now
    end
  end
  
  
  #
  # Utility Methods
  #
  
  
  # -------------------------
  # Readlog
  #
  def readlog
    db = YAML::Store.new('./yaml_db')
    db.transaction do
      @log = db['log']
    end
    @log = Array.new if @log.nil?
  end
  
  
  # -------------------------
  # Threads
  #
  def threads
    threadsSortedByNumber
  end
  
  
  # -------------------------
  # Threads sorted by number
  #
  def threadsSortedByNumber
    readlog if @log.nil?
    (1..@log.length-1).to_a
  end
  
  
  # -------------------------
  # Threads sorted by date
  #
  def threadsSortedByDate
    threadsSortedByNumber.sort{|a, b| @log[a]['LastModified'] <=> @log[b]['LastModified'] || a <=> b }
  end
  
  
  # -------------------------
  # Pager
  #
  def pager( current=nil )
    current = current.to_i
    result = { 'pagelist'=>[], 'current'=>current, 'later'=>nil, 'older'=>nil }
    
    #page表示調節
    pages = ( ( self.threads.length - 1 ) / @ThreadsPerPage ) + 1
    if current <= 1 then
      result['current'] = 1
      result['older'] = 2 if pages > 1
    elsif current >= pages then
      result['current'] = pages
      result['later'] = pages - 1 if pages > 1
    else
      result['later'] = current - 1
      result['older'] = current + 1
    end
    if pages < @MaxPages then
      result['pages'] = pages
      result['pagelist'] = (1..pages).to_a
      return result
    end
  end
  
  
  # -------------------------
  # RSS
  #
  def rss
    @rsp.content_type = 'application/rdf+xml; charset=utf-8'
    @rsp.write << strRSS
  end
  
  
  # -------------------------
  # Output RDF to File
  #
  def fileRDF
    rdf = File.open( @RDFFileName, IO::WRONLY|IO::CREAT )
    rdf.flock( IO::LOCK_EX )
    rdf.truncate( 0 )
    rdf << strRSS
    rdf.close
  end
  
  
  # -------------------------
  # Return RSS String
  #
  def strRSS
    @RSSMaxItem = 15
    latestArticles = Array.new
    result = ''
    
    countI = 0
    threadsSortedByDate.last(@RSSMaxItem).reverse.each do |n|
      thread = @log[n]
      countI += 1
      countJ = 0
      thread['children'].last( @RSSMaxItem - countI ).each do |article|
	countJ += 1
	item = Hash.new
	item['i'] = @log.index(thread)
	item['j'] = thread['children'].length - countJ
	item['body'] = article['body']
	item['name'] = article['name']
	item['time'] = article['time']
	item['time'] = article['time']
	item['_uri'] = '%s?read=%d#art%d-%d' % [ @URI, item['i'], item['i'], item['j'] ];
	if article.key?('subject') then
	  item['_subject'] = article['subject']
	elsif article.key?('title')
	  item['_subject'] = article['title']
	else
	  item['_subject'] = thread['subject']
	end
	latestArticles.push item
      end
    end
    latestArticles.sort!{|a,b|b['time']<=>a['time']}
    latestArticles.length = @RSSMaxItem if latestArticles.length > @RSSMaxItem
    
    result << <<"EOS"
<?xml version="1.0" encoding="utf-8" ?>
<rdf:RDF 
  xmlns="http://purl.org/rss/1.0/"
  xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#"
  xmlns:dc="http://purl.org/dc/elements/1.1/"
  xml:lang="ja">
 <channel rdf:about="#@URI">
  <title>#@Title</title>
  <link>#@URI</link>
  <description>#@PageTitle</description>
  <items>
   <rdf:Seq>
EOS
    latestArticles.each do |item|
      result << "  <rdf:li rdf:resource=\"%s\" />\n" % [ item['_uri'] ]
    end
    result << <<"EOS"
   </rdf:Seq>
  </items>
 </channel>

EOS
    titleRegexp = /#{RegexpUtf8Char}{0,90}/o
    descriptionRegexp = /#{RegexpUtf8Char}{0,450}/o
    latestArticles.each do |item|
      #本文の縮め処理
      uri = item['_uri']
      title = '%d-%d %s' % [ item['i'], item['j'], item['_subject'] ]
      title.gsub!(/&(\w*);/o,"&amp;$1;")
      title.slice!(90)
      title = title[titleRegexp]
      subject = item['_subject']
      subject.gsub!(/&(\w*);/o, "&amp;$1;")
      description = item['body']
      creator = item['name']
      creator.gsub!(/&(\w*);/o, "&amp;$1;")
      description.gsub!(/<BR\b[^>]*>/o,'　')
      description.gsub!(/<.*?>/o,'')
      description.gsub!(/&(\w*);/o, "&amp;$1;")
      description.slice!(450)
      description = description[descriptionRegexp]
      date = item['time'].getutc.strftime('%Y-%m-%dT%H:%M:%SZ')
      result << <<"EOS"
 <item rdf:about="#{uri}">
  <title>#{title}</title>
  <link>#{uri}</link>
  <description>#{description}</description>
  <dc:creator>#{creator}</dc:creator>
  <dc:subject>#{subject}</dc:subject>
  <dc:date>#{date}</dc:date>
 </item>

EOS
    end
    result << " </rdf:RDF>\n"
    return result
  end
end

#
# main
#
def main
  api = Web::Interface::AUTO.new
  api.each do |req|
    rsp = Web::Response.new
    begin
      sazaby = Sazaby.new( req, rsp )
      if req.key?('rss') || req['mode'] == 'rss' then
	sazaby.rss
      else
	updated = false
	if req['mode'] == 'postThread' then
	  sazaby.postThread
	  updated = true
	elsif req['mode'] == 'postArticle' then
	  sazaby.postArticle
	  updated = true
	end
	sazaby.view
	sazaby.fileRDF if updated
      end
    rescue => evar
      rsp << $!
      rsp << evar
    end
    api.response req,rsp
  end
end

# bootstrap
if __FILE__ == $0 then
  main
end
